@@ -15,11 +15,38 @@ use Scalar::Util qw(reftype blessed refaddr);
 use Config;
 use Carp qw(confess);
 use Storable qw(dclone);
+use Cwd;
 
 # Dynamically load constants from whatever is being tested
-our ($Class, $ConstClass);
+our ($Class, $ConstClass, $InRepo);
+sub get_git_top_dir {
+    my @dirs = (0, 1, 2, 4);
+    for my $d (@dirs) {
+        my $tdir = File::Spec->catdir(map File::Spec->updir, 1..$d);
+        my $gdir = File::Spec->catdir($tdir, '.git');
+        return $tdir
+            if -d $gdir;
+    }
+    return();
+}
+
+BEGIN{
+    if (defined(my $top_dir = get_git_top_dir())) {
+        for my $need ('Encoder', 'Decoder') {
+            my $blib_dir = File::Spec->catdir($top_dir, 'Perl', $need, "blib");
+            if (-d $blib_dir) {
+                require blib;
+                blib->import($blib_dir);
+            }
+        }
+        $InRepo=1;
+    }
+}
 BEGIN {
-    if (-e "lib/Sereal/Encoder.pm") {
+    if (-e "lib/Sereal.pm") {
+        $Class = 'Sereal::Encoder';
+    }
+    elsif (-e "lib/Sereal/Encoder.pm") {
         $Class = 'Sereal::Encoder';
     }
     elsif (-e "lib/Sereal/Decoder.pm") {
@@ -31,7 +58,7 @@ BEGIN {
     elsif (-e "lib/Sereal/Splitter.pm") {
         $Class = 'Sereal::Splitter';
     } else {
-        die "Could not find an applicable Sereal constants location";
+        die "Could not find an applicable Sereal constants location (in: ",cwd(),")";
     }
     $ConstClass = $Class . "::Constants";
     eval "use $ConstClass ':all'; 1"
@@ -601,16 +628,6 @@ sub setup_tests {
 }
 
 
-sub get_git_top_dir {
-    my @dirs = (0, 1, 2);
-    for my $d (@dirs) {
-        my $tdir = File::Spec->catdir(map File::Spec->updir, 1..$d);
-        my $gdir = File::Spec->catdir($tdir, '.git');
-        return $tdir
-            if -d $gdir;
-    }
-    return();
-}
 
 sub have_encoder_and_decoder {
     my ($min_v)= @_;
@@ -618,22 +635,15 @@ sub have_encoder_and_decoder {
     my $need = $Class =~ /Encoder/ ? "Decoder" : "Encoder";
     my $need_class = "Sereal::$need";
 
-    if (defined(my $top_dir = get_git_top_dir())) {
-        my $blib_dir = File::Spec->catdir($top_dir, 'Perl', $need, "blib");
-        if (-d $blib_dir) {
-            require blib;
-            blib->import($blib_dir);
-        }
-    }
     eval "use $Class; 1"
     or do {
-        diag("Could not locate $Class for testing" . ($@ ? " (Exception: $@)" : ""));
+        note("Could not locate $Class for testing" . ($@ ? " (Exception: $@)" : ""));
         return();
     };
 
     eval "use $need_class; 1"
     or do {
-        diag("Could not locate $need_class for testing" . ($@ ? " (Exception: $@)" : ""));
+        note("Could not locate $need_class for testing" . ($@ ? " (Exception: $@)" : ""));
         return();
     };
     my $cmp_v = $need_class->VERSION;
@@ -681,6 +691,10 @@ my $eng2= "1e3";
 
 my $sum= $eng0e0 + $eng0e1 + $eng2;
 
+sub encoder_required {
+    my ($ver, $name)= @_;
+    return "" . ( $Sereal::Encoder::VERSION < $ver ? "TODO " : "") . $name;
+}
 
 sub _get_roundtrip_tests {
     my @ScalarRoundtripTests = (
@@ -734,8 +748,8 @@ sub _get_roundtrip_tests {
             1023,1024,1025,
             8191,8192,8193,
         )),
-        ( map { [ ( $Sereal::Encoder::VERSION < 3.005002 ? "TODO" : "")
-                    . " troublesome num/strs '$_'", $_ ] } @numstr ),
+        ( map { [ encoder_required(3.005002, " troublesome num/strs '$_'"),
+                  $_ ] } @numstr ),
         ["long latin1 string", "üll" x 10000],
         ["long utf8 string", do {use utf8; " עדיין חשב" x 10000}],
         ["long utf8 string with only ascii", do {use utf8; "foo" x 10000}],
@@ -810,8 +824,27 @@ sub _get_roundtrip_tests {
         [ "undef", [\undef, \undef] ],
     );
 
+    my @blessed_array_check1;
+    $blessed_array_check1[0]= "foo";
+    $blessed_array_check1[1]= bless \$blessed_array_check1[0], "BlessedArrayCheck";
+    $blessed_array_check1[2]= \$blessed_array_check1[0];
+
+    my @blessed_array_check2= (3,0,0,3);
+    $blessed_array_check2[1]= \$blessed_array_check2[0];
+    $blessed_array_check2[2]= \$blessed_array_check2[3];
+    bless \$blessed_array_check2[0], "BlessedArrayCheck";
+    bless \$blessed_array_check2[3], "BlessedArrayCheck";
+
+    my @sc_array=(1,1);
+    $sc_array[0]=bless \$sc_array[1], "BlessedArrayLeft";
+    $sc_array[1]=bless \$sc_array[0], "BlessedArrayRight";
+
+
     my @RoundtripTests = (
         @ScalarRoundtripTests,
+        [ encoder_required(3.006006,"BlessedArrayCheck 1"), \@blessed_array_check1 ],
+        [ encoder_required(3.006006,"BlessedArrayCheck 2"), \@blessed_array_check2 ],
+        [ encoder_required(3.006006,"Scalar Cross Blessed Array"), \@sc_array ],
 
         ["[{foo => 1}, {foo => 2}] - repeated hash keys",
           [{foo => 1}, {foo => 2}] ],
@@ -1021,6 +1054,8 @@ sub run_roundtrip_tests_internal {
     my ($ename, $opt, $encode_decode_callbacks) = @_;
     require Data::Dumper;
 
+    my $failed = 0;
+
     my $decoder = Sereal::Decoder->new($opt);
     my $encoder = Sereal::Encoder->new($opt);
     my %seen_name;
@@ -1028,6 +1063,11 @@ sub run_roundtrip_tests_internal {
     foreach my $rt (@RoundtripTests) {
         my ($name, $data) = @$rt;
 
+        if ($failed > 20) {
+            fail("too many test failures to continue");
+            last;
+        }
+
         TODO:
         foreach my $meth (
               ['object-oriented',
@@ -1056,13 +1096,15 @@ sub run_roundtrip_tests_internal {
             eval {$encoded = $enc->($data); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while encoding: $err");
+                    fail("$name ($ename, $mname, encoding failed)");
+                    $failed++;
                 };
 
             defined($encoded)
                 or do {
                     fail("$name ($ename, $mname, encoded defined)");
                     debug_checks(\$data, \$encoded, undef);
+                    $failed++;
                     next; #test
                 };
 
@@ -1070,13 +1112,16 @@ sub run_roundtrip_tests_internal {
             eval {$decoded = $dec->($encoded); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while decoding: $err");
+                    fail("$name ($ename, $mname, decoding failed)");
+                    $failed++;
+                    next;
                 };
 
             defined($decoded) == defined($data)
                 or do {
                     fail("$name ($ename, $mname, decoded definedness)");
                     debug_checks(\$data, \$encoded, undef);
+                    $failed++;
                     next; #test
                 };
 
@@ -1085,13 +1130,15 @@ sub run_roundtrip_tests_internal {
             eval {$encoded2 = $enc->($decoded); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while encoding the second time: $err");
+                    fail("$name ($ename, $mname, second encoding failed)");
+                    $failed++;
+                    next; #test
                 };
 
             defined $encoded2
                 or do {
                     fail("$name ($ename, $mname, encoded2 defined)");
-                    debug_checks(\$data, \$encoded, \$decoded);
+                    $failed++;
                     next; #test
                 };
 
@@ -1099,13 +1146,15 @@ sub run_roundtrip_tests_internal {
             eval {$decoded2 = $dec->($encoded2); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while decoding the second time: $err");
-                    # hobodecode($encoded2);
+                    fail("$name ($ename, $mname, second decoding failed)");
+                    $failed++;
+                    next; #test
                 };
 
             defined($decoded2) == defined($data)
                 or do {
                     fail("$name ($ename, $mname, decoded2 defined)");
+                    $failed++;
                     next; #test
                 };
 
@@ -1114,13 +1163,15 @@ sub run_roundtrip_tests_internal {
             eval {$encoded3 = $enc->($decoded2); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while encoding the third time: $err");
+                    fail("$name ($ename, $mname, third encoding failed)");
+                    $failed++;
+                    next; #test
                 };
 
             defined $encoded3
                 or do {
                     fail("$name ($ename, $mname, encoded3 defined)");
-                    debug_checks(\$data, \$encoded, \$decoded);
+                    $failed++;
                     next; #test
                 };
 
@@ -1128,31 +1179,38 @@ sub run_roundtrip_tests_internal {
             eval {$decoded3 = $dec->($encoded3); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while decoding the third time: $err");
+                    fail("$name ($ename, $mname, third decoding failed)");
+                    $failed++;
+                    next; #test
                 };
 
             defined($decoded3) == defined($data)
                 or do {
                     fail("$name ($ename, $mname, decoded3 defined)");
+                    $failed++;
                     next; #test
                 };
 
-            deep_cmp($decoded, $data,       "$name ($ename, $mname, decoded vs data)") or next; #test
-            deep_cmp($decoded2, $data,      "$name ($ename, $mname, decoded2 vs data)") or next; #test
-            deep_cmp($decoded2, $decoded,   "$name ($ename, $mname, decoded2 vs decoded)") or next; #test
-
-            deep_cmp($decoded3, $data,      "$name ($ename, $mname, decoded3 vs data)") or next; #test
-            deep_cmp($decoded3, $decoded,   "$name ($ename, $mname, decoded3 vs decoded)") or next; #test
-            deep_cmp($decoded3, $decoded2,  "$name ($ename, $mname, decoded3 vs decoded2)") or next; #test
+            deep_cmp($decoded, $data,       "$name ($ename, $mname, decoded vs data)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded2, $data,      "$name ($ename, $mname, decoded2 vs data)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded2, $decoded,   "$name ($ename, $mname, decoded2 vs decoded)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded3, $data,      "$name ($ename, $mname, decoded3 vs data)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded3, $decoded,   "$name ($ename, $mname, decoded3 vs decoded)")
+                or do { $failed++; next }; #test
+            deep_cmp($decoded3, $decoded2,  "$name ($ename, $mname, decoded3 vs decoded2)")
+                or do { $failed++; next }; #test
 
             if ( $ename =~ /canon/ ) {
-                deep_cmp($encoded2, $encoded,  "$name ($ename, $mname, encoded2 vs encoded)") or do {
-                    diag Dumper($encoded2);
-                    diag Dumper($encoded);
-                    next; #test
-                };
-                deep_cmp($encoded3, $encoded2, "$name ($ename, $mname, encoded3 vs encoded2)") or next; #test
-                deep_cmp($encoded3, $encoded,  "$name ($ename, $mname, encoded3 vs encoded)") or next; #test
+                deep_cmp($encoded2, $encoded,  "$name ($ename, $mname, encoded2 vs encoded)")
+                    or do { $failed++; next }; #test
+                deep_cmp($encoded3, $encoded2, "$name ($ename, $mname, encoded3 vs encoded2)")
+                    or do { $failed++; next }; #test
+                deep_cmp($encoded3, $encoded,  "$name ($ename, $mname, encoded3 vs encoded)")
+                    or do { $failed++; next }; #test
 
                 if ($ENV{SEREAL_TEST_SAVE_OUTPUT} and $mname eq 'object-oriented') {
                     use File::Path;
@@ -1161,8 +1219,14 @@ sub run_roundtrip_tests_internal {
                         my @clean= ($ename, $name);
                         s/[^\w.-]+/_/g, s/__+/_/g for @clean;
                         my $cleaned= join "/", @clean;
-                        my $dir= $0;
-                        $dir=~s!/[^/]+\z!/data/$clean[0]!;
+                        my ($v,$p,$d)= File::Spec->splitpath($0);
+                        my $dir= File::Spec->catpath(
+                            $v, 
+                            File::Spec->catdir(
+                                File::Spec->splitdir($p),
+                                "data",$clean[0]
+                            )
+                        );
                         mkpath $dir unless -d $dir;
                         my $base= "$dir/$clean[1].enc";
                         $seen_name{$combined_name}= $base;
@@ -1198,13 +1262,13 @@ sub _write_file {
 # For bootstrapping other language implementations' tests
 our $COMPRESS;
 sub write_test_files {
-    my $dir = shift;
+    my ($dir, $version) = @_;
     require File::Path;
     File::Path::mkpath($dir);
     my $make_data_file_name = sub {File::Spec->catfile($dir, sprintf("test_data_%05u", shift))};
     my $make_name_file_name = sub {File::Spec->catfile($dir, sprintf("test_name_%05u", shift))};
 
-    setup_tests();
+    setup_tests($version);
     foreach my $testno (1..@BasicTests) {
         my $t = $BasicTests[$testno-1];
         my $data = ref($t->[1]) eq 'CODE' ? $t->[1]->() : $t->[1];
